# Přístupová práva

from enum import Enum, auto
from dataclasses import dataclass
from typing import Set, List, Dict, Tuple, Optional

import mo
import mo.db as db


class Right(Enum):
    assign_rights = auto()
    edit_region = auto()
    edit_place = auto()
    manage_round = auto()
    manage_contest = auto()
    add_contest = auto()
    view_contestants = auto()       # Prohlížet si seznam účastníků
    upload_solutions = auto()       # Odevzdávat za účastníka ve stavu "running"
    upload_feedback = auto()        # Nahrávat opravené řešení ve stavu "grading"
    view_submits = auto()           # Prohlížet si řešení a opravy
    upload_submits = auto()         # Nahrávat řešení a opravy nezávisle na stavu soutěže
    edit_points = auto()            # Přidělovat body ve stavu "grading"
    view_statement = auto()         # Prohlížet zadání, pokud je dostupné pro dozor
    add_users = auto()
    edit_users = auto()
    add_orgs = auto()
    edit_orgs = auto()


@dataclass
class Role:
    role: db.RoleType
    name: str
    rights: Set[Right]

    def __init__(self, role: db.RoleType, rights: Set[Right]):
        self.role = role
        self.name = role.friendly_name()
        self.rights = rights


# Order in this list represents hierarchy for assign_rights right
# (garant could assign role of garant or garant_kraj, but garant_kraj cannot
# assign role garant).
roles: List[Role] = [
    Role(
        role=db.RoleType.garant,
        rights={
            Right.assign_rights,
            Right.edit_place,
            Right.manage_round,
            Right.manage_contest,
            Right.add_contest,
            Right.view_contestants,
            Right.view_submits,
            Right.upload_submits,
            Right.edit_points,
            Right.add_users,
            Right.edit_users,
            Right.add_orgs,
            Right.edit_orgs,
        },
    ),
    Role(
        role=db.RoleType.garant_kraj,
        rights={
            Right.assign_rights,
            Right.edit_place,
            Right.manage_contest,
            Right.add_contest,
            Right.view_contestants,
            Right.view_submits,
            Right.upload_submits,
            Right.edit_points,
            Right.view_statement,
            Right.add_users,
            Right.edit_users,
            Right.add_orgs,
            Right.edit_orgs,
        },
    ),
    Role(
        role=db.RoleType.garant_okres,
        rights={
            Right.assign_rights,
            Right.edit_place,
            Right.manage_contest,
            Right.add_contest,
            Right.view_contestants,
            Right.view_submits,
            Right.upload_submits,
            Right.edit_points,
            Right.view_statement,
            Right.add_users,
            Right.edit_users,
            Right.add_orgs,
            Right.edit_orgs,
        },
    ),
    Role(
        role=db.RoleType.garant_skola,
        rights={
            # FIXME: Až se pořádně rozjedou školní kola, asi chceme školním správcům omezit
            # práva na editaci uživatelů. Viz issue #66.
            Right.assign_rights,
            Right.edit_place,
            Right.manage_contest,
            Right.view_contestants,
            Right.view_submits,
            Right.upload_submits,
            Right.edit_points,
            Right.view_statement,
            Right.add_users,
            Right.edit_users,
            Right.add_orgs,
            Right.edit_orgs,
        },
    ),
    Role(
        role=db.RoleType.dozor,
        rights={
            Right.view_contestants,
            Right.view_submits,
            Right.upload_solutions,
            Right.view_statement,
        },
    ),
    Role(
        role=db.RoleType.opravovatel,
        rights={
            Right.view_contestants,
            Right.view_submits,
            Right.upload_feedback,
            Right.edit_points,
            Right.view_statement,
        },
    ),
]


roles_by_type: Dict[db.RoleType, Role] = {r.role: r for r in roles}
role_order_by_type: Dict[db.RoleType, int] = {r.role: i for (i, r) in enumerate(roles)}


class Rights:
    """Třída, která popisuje množinu práv k nějakému objektu."""

    user: db.User
    user_roles: List[db.UserRole]
    rights: Set[Right]

    def __repr__(self):
        ros = " ".join([r.role.name for r in self.user_roles])
        ris = " ".join([r.name for r in self.rights])
        return f"Rights(uid={self.user.user_id} is_admin={self.user.is_admin} roles=<{ros}> rights=<{ris}>)"

    def _clone_from(self, src: 'Rights'):
        self.user = src.user
        self.user_roles = src.user_roles
        self.rights = src.rights

    def have_right(self, right: Right) -> bool:
        if self.user.is_admin:
            return True
        assert self.rights is not None
        return right in self.rights

    def get_roles(self) -> List[db.RoleType]:
        """Seznam unikátních rolí, stříděných od nejvýznamnější."""
        role_set = set(ur.role for ur in self.user_roles)
        return sorted(role_set, key=lambda r: role_order_by_type[r])

    # Práva na práci s podřízenými místy

    def can_edit_place(self, place: db.Place) -> bool:
        if self.have_right(Right.edit_region):
            return True
        elif place.level >= 3 and self.have_right(Right.edit_place):
            # level >= 3 ... city and lower
            return True
        return False

    def can_add_place_child(self, place: db.Place) -> bool:
        if self.have_right(Right.edit_region):
            return True
        elif place.level >= 2 and self.have_right(Right.edit_place):
            # Can add cities and lower
            return True
        return False

    # Práva na práci s uživateli

    def can_edit_user(self, user: db.User) -> bool:
        if user.is_admin:
            return self.user.is_admin  # only admins can edit admins
        elif user.is_org:
            return self.have_right(Right.edit_orgs)
        return self.have_right(Right.edit_users)

    # Interní rozhodovaní o dostupnosti zadání

    def _check_view_statement(self, round: db.Round):
        if round.tasks_file is None:
            return False

        if self.have_right(Right.manage_round):
            # Správce kola může vždy všechno
            return True

        # Pokud už soutěž skončila, přístup k zadání má každý org.
        # XXX: Rozhodujeme podle stavu kola, nikoliv soutěže!
        if round.state in [db.RoundState.grading, db.RoundState.closed]:
            return True

        # Od stanoveného času vidí zadání orgové s právem view_statement.
        if (self.have_right(Right.view_statement)
                and round.state != db.RoundState.preparing
                and round.pr_tasks_start is not None
                and mo.now >= round.pr_tasks_start):
            return True

        # Ve zbylých případech jsme konzervativní a zadání neukazujeme
        return False


class RoundRights(Rights):
    """Práva ke kolu."""

    round: db.Round

    def __repr__(self):
        ros = " ".join([r.role.name for r in self.user_roles])
        ris = " ".join([r.name for r in self.rights])
        return f"RoundRights(uid={self.user.user_id} is_admin={self.user.is_admin} round=#{self.round.round_id} roles=<{ros}> rights=<{ris}>)"

    # Metody offer_* testují, zda se má v UI nabízet příslušná funkce.
    # Skutečnou kontrolu práv dělá až implementace funkce podle stavu soutěže.

    def _is_active(self) -> bool:
        return self.round.state not in [db.RoundState.preparing, db.RoundState.closed]

    def offer_upload_solutions(self) -> bool:
        return (self.have_right(Right.upload_submits)
                or (self.have_right(Right.upload_solutions) and self._is_active()))

    def offer_upload_feedback(self) -> bool:
        return (self.have_right(Right.upload_submits)
                or (self.have_right(Right.upload_feedback) and self._is_active()))

    def offer_edit_points(self) -> bool:
        return (self.have_right(Right.manage_contest)
                or (self.have_right(Right.edit_points) and self._is_active()))

    def can_view_statement(self) -> bool:
        return self._check_view_statement(self.round)


class ContestRights(Rights):
    """Práva k soutěži."""

    contest: db.Contest

    def __repr__(self):
        ros = " ".join([r.role.name for r in self.user_roles])
        ris = " ".join([r.name for r in self.rights])
        return f"ContestRights(uid={self.user.user_id} is_admin={self.user.is_admin} contest=#{self.contest.contest_id} roles=<{ros}> rights=<{ris}>)"

    def can_upload_solutions(self) -> bool:
        return (self.have_right(Right.upload_submits)
                or self.have_right(Right.upload_solutions) and self.contest.state == db.RoundState.running)

    def can_upload_feedback(self) -> bool:
        return (self.have_right(Right.upload_submits)
                or self.have_right(Right.upload_feedback) and self.contest.state == db.RoundState.grading)

    def can_edit_points(self) -> bool:
        return (self.have_right(Right.edit_points) and self.contest.state == db.RoundState.grading
                or self.have_right(Right.manage_contest))

    def can_view_statement(self) -> bool:
        return self._check_view_statement(self.contest.round)


class Gatekeeper:
    """Dveřník rozhoduje, jaká práva má uživatel k danému objektu.
    Typicky existuje jen v jedné instanci, navázaná na aktuální request
    nebo právě zpracovávaný job."""

    user: db.User
    roles: List[db.UserRole]
    parent_cache: Dict[int, List[db.Place]]
    rights_cache: Dict[Tuple[Optional[int], Optional[int], Optional[str], Optional[int], Optional[db.RoleType]], Rights]

    def __init__(self, user: db.User):
        self.user = user
        self.roles = user.roles
        assert user.is_org or user.is_admin
        self.parent_cache = {}
        self.rights_cache = {}

    def get_parents(self, place: db.Place) -> List[db.Place]:
        pid = place.place_id
        if pid not in self.parent_cache:
            self.parent_cache[pid] = db.get_place_parents(place)
        return self.parent_cache[pid]

    def rights_for(
            self, place: Optional[db.Place] = None, year: Optional[int] = None,
            cat: Optional[str] = None, seq: Optional[int] = None,
            min_role: Optional[db.RoleType] = None) -> Rights:
        """Posbírá role a práva, která se vztahují k danému místu (možno i tranzitivně) a soutěži.
        Pokud place=None, omezení role na místo se nebere v úvahu.
        Pokud year==None, vyhovují role s libovolným ročníkem; pokud year=0, vyhovují jen ty s neuvedeným ročníkem.
        Podobně cat a seq.
        Pokud min_role!=None, tak se uvažují jen role, které jsou v hierarchii alespoň na úrovni min_role."""

        cache_key = (place.place_id if place is not None else None, year, cat, seq, min_role)
        if cache_key in self.rights_cache:
            return self.rights_cache[cache_key]

        rights = Rights()
        rights.user = self.user
        rights.user_roles = []
        rights.rights = set()

        def try_role(role: db.UserRole, at: Optional[db.Place]):
            if ((at is None or role.place_id == at.place_id)
                    and (role.year is None or year is None or role.year == year)
                    and (role.category is None or cat is None or role.category == cat or (role.category == 'Z' and cat.startswith('Z')))
                    and (role.seq is None or seq is None or role.seq == seq)
                    and (min_role is None or role_order_by_type[min_role] >= role_order_by_type[role.role])):
                rights.user_roles.append(role)
                r = roles_by_type[role.role]
                rights.rights |= r.rights

        # XXX: This might be faster...
        if place is None:
            for role in self.roles:
                try_role(role, None)
        else:
            for at in self.get_parents(place):
                for role in self.roles:
                    try_role(role, at)

        self.rights_cache[cache_key] = rights
        return rights

    def rights_generic(self):
        """Posbírá role a práva, ale ignoruje omezení rolí na místa a soutěže. Hodí se pro práva k editaci uživatelů apod."""
        return self.rights_for()

    def rights_for_round(self, round: db.Round, any_place: bool) -> RoundRights:
        if any_place:
            place = None
        else:
            place = db.get_root_place()
        rights = RoundRights()
        rights.round = round
        rights._clone_from(self.rights_for(
            place=place,
            year=round.year,
            cat=round.category,
            seq=round.seq,
        ))
        return rights

    def rights_for_contest(self, contest: db.Contest, site: Optional[db.Place] = None) -> ContestRights:
        rights = ContestRights()
        rights.contest = contest
        rights._clone_from(self.rights_for(
            place=site or contest.place,
            year=contest.round.year,
            cat=contest.round.category,
            seq=contest.round.seq,
        ))
        return rights

    def can_set_role(self, role: db.UserRole):
        # We will use rights_for but we need to slightly modify its behavior. Can
        # set role with * year only if current user has role with * year, thus we
        # need to pass year=0 for * year (cat and seq similarly).
        rights = self.rights_for(
            place=role.place,
            year=role.year or 0,
            cat=role.category or '',
            seq=role.seq or 0,
            min_role=role.role
        )
        return rights.have_right(Right.assign_rights)
